//****************************************************************************************
//
//
//                                 SOUND  DAC  LIBRARY
//
//
//  Library developed by Noé Paniagua and Kiki Gómez, University Of Málaga (Electronica Creativita students).
//
//  Patron: DIY synth © 2024 by Noé Paniagua, Juan Carlos Gómez is licensed under CC BY-NC 4.0 
//
//**************************************************************************************
#include "FspTimer.h"

//-------------------------------------
//    SYNTH HARDWARE CONFIGURATIONS
//-------------------------------------
#define NUMBER_OF_KEYS 13
#define DAC_RESOLUTION 1024

//-------------------------------
//    SYNTH SOFTWARE VARIABLES
//-------------------------------
#define NUMBER_OF_FRECS 39
static uint16_t frecuenciasAfinadas[NUMBER_OF_FRECS] = {130, 138, 146, 155, 164, 174, 185, 196, 207, 220, 233, 246, 261,
                                                        277, 293, 311, 329, 349, 370, 392, 415, 440, 466, 493, 523, 554,
                                                        587, 622, 659, 698, 740, 784, 830, 880, 932, 987, 1046, 1108, 1174};
uint8_t octaveCount = 0;
static volatile uint16_t NUMBER_OF_SAMPLES[NUMBER_OF_KEYS];//number of samples: calculate using sample time and frecuency

static volatile uint32_t samplerFrecuency = 40000;
static volatile uint16_t SIGNAL_MAX_SIZE = 1024;  // hardware max value
static volatile uint16_t SIGNAL_MAX_SIZE_TRI = 2048;  // hardware max value
static volatile uint16_t newSIGNAL_MAX_SIZE = 1024; // hardware max readed value
static volatile uint16_t LOCAL_SIGNAL_SIZE = 1024;// max value for each key (then, change 4 by countKeyPressed)

static volatile uint8_t kindOfWave = 0;
static volatile uint8_t keys[NUMBER_OF_KEYS] = {0,0,0,0};
static volatile uint8_t lastKeys[NUMBER_OF_KEYS] = {0,0,0,0};
static uint8_t countKeysPressed;

static uint16_t currentFrecuency = 0;
static uint8_t filtroFrecuency = 0;
float volume = 1;

//*********ADSR VARIABLES**********
#define MAX_TIME 2

typedef struct ADSR{
  uint8_t semAttack;
  uint8_t semDecay;
  uint8_t semSustain;
  uint8_t semRelease;
  uint16_t timingHighFrec;
  float mod;
};

static ADSR adsr[NUMBER_OF_KEYS];

typedef struct ENVELOPE{
  float resolution;
};
static ENVELOPE envelopeAttack;
static ENVELOPE envelopeDecay;
static float envelopeSustainMod;
static ENVELOPE envelopeRelease;

//******** WAVEFORM VARIABLES**********
static float frecuencyResolution[NUMBER_OF_KEYS];
static float outputAmplitude[NUMBER_OF_KEYS];

float mapf(float x, float in_min, float in_max, float out_min, float out_max)
{
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

//********* KEYS LOGIC VARIABLES**********
bool array_teclas[32]; 
bool teclas[12];
unsigned long tiempo = 0; 
uint8_t cuenta = 0;

//-------------------------------------
//    KEYS LOGIC READ FUNCTION
//-------------------------------------
void synthKeysLogic(void){
  if(millis() >tiempo + keysSamplingPeriod){
    digitalWrite(SELECTOR0,cuenta&1);
    digitalWrite(SELECTOR1,(cuenta&2)>>1);
    digitalWrite(SELECTOR2,(cuenta&4)>>2);

    bool entrada1 = digitalRead(SALIDA1_TECLAS);
    bool entrada2 = digitalRead(SALIDA2_TECLAS);
    //Serial.println(entrada2);
    

    array_teclas[cuenta*2] = entrada1;
    array_teclas[(cuenta*2)+1] = entrada2;

    tiempo = millis();

    if(cuenta == 15){
      cuenta = 0;
      teclas[0] = !array_teclas[9];
      teclas[1] = !array_teclas[11];
      teclas[2] = !array_teclas[1];
      teclas[3] = !array_teclas[3];
      teclas[4] = array_teclas[2];
      teclas[5] = array_teclas[8];
      teclas[6] = array_teclas[10];
      teclas[7] = array_teclas[14];
      teclas[8] = array_teclas[12];
      teclas[9] = !array_teclas[5];
      teclas[10] = array_teclas[4];
      teclas[11] = array_teclas[6];
      teclas[12] = array_teclas[0];

      for(int m=0; m < 13; m++){
        synthKeysState(m, teclas[m]);
      }
    }
    else
      cuenta++;
  }
}

//-------------------------------------
//    N-KEY STATE PRESSED/UNPRESSED
//-------------------------------------
void synthKeysState(uint8_t pressedKey, uint8_t keyState){
  keys[pressedKey] = keyState;

  if(lastKeys[pressedKey] != keys[pressedKey]){
    if(keys[pressedKey] == 1){ 
      adsr[pressedKey].semAttack = 1; 
      adsr[pressedKey].mod = 0;
      outputAmplitude[pressedKey] = 0;
    }
    else {
      adsr[pressedKey].semRelease = 1; 
    }
  }
  lastKeys[pressedKey] = keys[pressedKey];
}

//-------------------------------------
//    KEYS PRESSED COUNTER
//-------------------------------------
void synthKeysCounter(void){
  uint8_t TEMP_countKeysPressed = 0;
   for(uint8_t i = 0; i < NUMBER_OF_KEYS; i++)
     if(keys[i] || adsr[i].semRelease) TEMP_countKeysPressed++; 

    if(!TEMP_countKeysPressed)
      countKeysPressed = 1;// semi atomic!!
    else
      countKeysPressed = TEMP_countKeysPressed;
}

//-------------------
//    ADSR REFRESH
//-------------------
void synthADSR(uint16_t AttackPot, uint16_t DecayPot, uint16_t SustainPot, uint16_t ReleasePot){
  float tempTime, nSteps = 0;
  
  //
  //ATTACK
  //
  tempTime = mapf(AttackPot, 0, 1024, 0.1, MAX_TIME);
  nSteps = (float)tempTime / (float)(0.01);
  envelopeAttack.resolution = (float)1/(float)(nSteps);
  
  //
  //SUSTAIN
  //
  envelopeSustainMod = mapf(SustainPot, 0 , 1024, 0.1, 1);

  //
  //DECAY
  //
  tempTime = mapf(DecayPot, 0, 1024, 0.1, MAX_TIME);
  nSteps = (float)tempTime / (float)(0.01);
  envelopeDecay.resolution = (1-envelopeSustainMod)/(float)(nSteps);
  
  //
  //RELEASE
  //
  tempTime = mapf(ReleasePot, 0, 1024, 0.1, MAX_TIME);
  nSteps = (float)tempTime / (float)(0.01);
  envelopeRelease.resolution = (envelopeSustainMod)/(float)(nSteps);
}

//--------------------------
//    GOOD FRECUENCY
//--------------------------
void setNewFrecuency(uint16_t octave){
  float period, nSteps, resolution, frecuency;
  uint8_t frecPos = 0 + ((octave-1)*13);
  
  for(uint8_t i = 0; i < NUMBER_OF_KEYS; i++){
    frecuency = frecuenciasAfinadas[frecPos+i];
    resolution = (float)1/400;
    period = 1/frecuency;
    period *= 100;
  
    if(kindOfWave == 0){
      period = period /2;
      nSteps = (float)period / (float)(resolution);
      frecuencyResolution[i] = (SIGNAL_MAX_SIZE)/(float)(nSteps);
    }
    else if(kindOfWave == 1){//Square Wave
      nSteps = (float)period / (float)(resolution);
      frecuencyResolution[i] = (SIGNAL_MAX_SIZE)/(float)(nSteps);
    }

    //
    // Saw-tooth waveform
    else if(kindOfWave == 2){
      nSteps = (float)period / (float)(resolution);
      frecuencyResolution[i] = (SIGNAL_MAX_SIZE_TRI)/(float)(nSteps);
    }
  }
}

//--------------------------
//    SET SYTH MAX VOLUME 
//---------------------------
void synthSetVolume(int newVolume){
  volume = mapf(newVolume, 0, 1024, 0, 1);
}

//-----------------------------------
//    DIFFERENTS WAVEFORMS SETS
//
//    [0]-> Triangle    [1]-> Square
//    [2]-> SawTooth    [3]-> XD(sine uwu)NOT YET
//-----------------------------------
void synthSetWaveForm(uint8_t wave){
  kindOfWave = wave;
}

//---------------------------------
//  CALLBACK ISR USED BY TIMER
//---------------------------------
void timer_callback(timer_callback_args_t __attribute((unused)) *p_args) {
  uint16_t lastSample[NUMBER_OF_KEYS], finalSample = 0, nextSample, subLooperTimer, looperTemp;

  for(uint8_t i = 0; i < NUMBER_OF_KEYS; i++){  
      
    if(keys[i] || adsr[i].semRelease){// if clicked...
      //
      //HUGE TIMER
      //
      if(adsr[i].timingHighFrec++ == 400){//10ms
         adsr[i].timingHighFrec = 0;

         //
         // ATTACK
         //
         if(adsr[i].semAttack){
           adsr[i].mod += envelopeAttack.resolution;
           
           if(adsr[i].mod >= 1){//finish attack
            adsr[i].semAttack = 0;
            adsr[i].semDecay = 1;
            adsr[i].mod = 1;
           }
         }

         //
         // DECAY
         //
         else if(adsr[i].semDecay){
          adsr[i].mod -= envelopeDecay.resolution;
          
          if(adsr[i].mod <= envelopeSustainMod){//finish decay
            adsr[i].semDecay = 0;
            adsr[i].mod = envelopeSustainMod;
          }
        }
        
        //
        // RELEASE
        // 
        else if(adsr[i].semRelease){//if release
          adsr[i].mod -= envelopeRelease.resolution;
          
          if(adsr[i].mod <= 0){//finish decay
            adsr[i].semRelease = 0;
            adsr[i].mod = 0;
          }
        }    
      }//end huge timer


      //
      // SET THE WAVEFORMMMM
      //
      outputAmplitude[i] += frecuencyResolution[i];

      // 
      // TRIANGLE
      //
      if(kindOfWave == 0){
        if(outputAmplitude[i] >= SIGNAL_MAX_SIZE_TRI<<1){
          SIGNAL_MAX_SIZE_TRI = newSIGNAL_MAX_SIZE<<1;
          outputAmplitude[i] = 0;
        }
        if(outputAmplitude[i] <= SIGNAL_MAX_SIZE_TRI)
          nextSample = outputAmplitude[i];
        else
          nextSample = (SIGNAL_MAX_SIZE_TRI<<1) - outputAmplitude[i];
      }

      // 
      // SQUARE
      else if(kindOfWave == 1){//Square Wave
        if(outputAmplitude[i] >= SIGNAL_MAX_SIZE){
          SIGNAL_MAX_SIZE = newSIGNAL_MAX_SIZE;
          outputAmplitude[i] = 0;
        }
       if(outputAmplitude[i] >= SIGNAL_MAX_SIZE>>1) nextSample = SIGNAL_MAX_SIZE;
       else nextSample = 0;
      }

      // 
      // SAWTOOTH
      //
      else if(kindOfWave == 2){
        if(outputAmplitude[i] >= LOCAL_SIGNAL_SIZE){
          SIGNAL_MAX_SIZE = newSIGNAL_MAX_SIZE;
          outputAmplitude[i] = 0;
        }
        nextSample = outputAmplitude[i];
      }

      lastSample[i] = (uint16_t)(nextSample*(float)adsr[i].mod * (float)volume);
    }
    else lastSample[i] = 0;

    finalSample += lastSample[i];
  }// for all keys[i]
    
  *DAC12_DADR0 = finalSample;  
} 

//---------------------------------
//      SAMPLER TIMER SETUP
//---------------------------------
bool synthBeginTimer(float rate) {
  uint8_t timer_type = GPT_TIMER;
  int8_t tindex = FspTimer::get_available_timer(timer_type);

  samplerFrecuency = rate;
  
  if (tindex < 0){
    tindex = FspTimer::get_available_timer(timer_type, true);
  }
  if (tindex < 0){
    return false;
  }
  FspTimer::force_use_of_pwm_reserved_timer();
  if(!audio_timer.begin(TIMER_MODE_PERIODIC, timer_type, tindex, rate, 0.0f, timer_callback)){
    return false;
  }
  if (!audio_timer.setup_overflow_irq()){
    return false;
  }
  if (!audio_timer.open()){
    return false;
  }
  if (!audio_timer.start()){
    return false;
  }
  return true;
}

//---------------------------------
//SETUP ARDUINO R4 DAC--> PIN A0
//---------------------------------
void synthSetupDac(void)       // Note make sure ADC is stopped before setup DAC
  {
  *MSTP_MSTPCRD &= ~(0x01 << MSTPD20);  // Enable DAC12 module
  *DAC12_DADPR    = 0x00;               // DADR0 Format Select Register - Set right-justified format
//  *DAC12_DAADSCR  = 0x80;             // D/A A/D Synchronous Start Control Register - Enable
  *DAC12_DAADSCR  = 0x00;               // D/A A/D Synchronous Start Control Register - Default
// 36.3.2 Notes on Using the Internal Reference Voltage as the Reference Voltage
  *DAC12_DAVREFCR = 0x00;               // D/A VREF Control Register - Write 0x00 first - see 36.2.5
  *DAC12_DADR0    = 0x0000;             // D/A Data Register 0 
   delayMicroseconds(10);               // Needed delay - see data sheet
  *DAC12_DAVREFCR = 0x01;               // D/A VREF Control Register - Select AVCC0/AVSS0 for Vref
  *DAC12_DACR     = 0x5F;               // D/A Control Register - 
   delayMicroseconds(5);                // Needed delay - see data sheet
  *DAC12_DADR0    = 2048;               // D/A Data Register 0 - value of mid range bias  
  *PFS_P014PFS   = 0x00000000;          // Port Mode Control - Make sure all bits cleared
  *PFS_P014PFS  |= (0x1 << 15);         // ... use as an analog pin
  }
